#Příliš žluťoučký kůň úpěl ďábelské ó - PŘÍLIŠ ŽLUŤOUČKÝ KŮŇ ÚPĚL ĎÁBELSKÉ Ó
#R:/p1_PGM/p1_INP/game/db/__init__.py
"""
Šablona initorů balíčků.
"""
import dbg; dbg.start_pkg(0, __name__, __doc__)
###########################################################################q

from dataclasses            import *
from game.api.scen_types    import *
from game.api.scenario      import ScenarioStep



###########################################################################q
# Pomocné funkce

# @dbg.prSEdr()
def id_game(): id_game.id += 1;   return id_game.id
id_game.id = 0

# @dbg.prSEdr()
def id_action(): id_action.id += 1; return id_action.id
id_action.id = 0

# @dbg.prSEdr()
def id_place(): id_place.id += 1;   return id_place.id
id_place.id = 0

# @dbg.prSEdr()
def id_item(): id_item.id += 1;   return id_item.id
id_item.id = 0



###########################################################################q
# Typy záznamů v jednotlivých tabulkách databáze

@dataclass
class Rec_Game:
    """Záznam o hře s názvem, popisem a odkazem na autora."""
    Gm_Author_ID    :str
    Gm_Name         :str
    Gm_Description  :str
    Gm_ID           :int = field(default_factory=id_game)


@dataclass
class Rec_Action:
    """Záznam o akci hry s typem, názvem, popisem a odpovědí."""
    Ac_Game     :int
    Ac_Type     :str
    Ac_Name     :str
    Ac_Answer   :str
    Ac_ID       :int = field(default_factory=id_action)


@dataclass
class Rec_Place:
    """Záznam o prostoru hry s jeho názvem a popisem."""
    Pl_Game         :int
    Pl_Name         :str
    Pl_Description  :str
    Pl_ID           :int = field(default_factory=id_place)


@dataclass
class Rec_Item:
    """Záznam o h-objektu hry s jeho názvem a vahou."""
    It_Place    :int
    It_Name     :str
    It_Weight   :int
    It_ID       :int = field(default_factory=id_item)


@dataclass
class Rec_Neighbor:
    """Záznamy o vzájemném sousedství prostorů. Záznam oznamuje,
    že z prostoru Source se lze přesunout do prostoru Dest.
    """
    Ne_Source   :int
    Ne_Dest     :int



###########################################################################q

class Table:
    """Tabulka databáze"""

    def __init__(self, name:str, field_names):
        """Vytvoří prázdnou tabulku zadaného názvu se zadanými názvy polí.
        Názvy polí je možno zadat jako seznam stringů nebo jako string
        s jednotlivými názvy oddělenými bílými znaky.
        """
        if isinstance(field_names, str):
            field_names = field_names.split()
        self.name = name
        self.field_names = field_names
        self.records = []

    def add(self, record):
        """Přidání záznamu do tabulky."""
        self.records.append(record)

    def __str__(self):
        result = [f'Tabulka: {self.name}',
                  f'Pole:    {self.field_names}']
        for record in self.records:
            result.append(str(record))
        return '\n'.join(result)


# Vytvoření potřebných tabulek
tbl_Games    = Table('Games',   'Gm_ID Gm_Name Gm_Author_ID Gm_Description')
tbl_Actions  = Table('Actions', 'Ac_ID Ac_Game Ac_Type Ac_Name '
                                'Ac_Description Ac_Answer')
tbl_Places   = Table('Places',  'Pl_ID Pl_Game Pl_Name Pl_Description')
tbl_Items    = Table('Items',   'It_ID It_Game It_Name It_Weight')
tbl_Neighbors= Table('Neighbors','Ne_Source, Ne_Dest')

# Používané typy kroků
checked_types:set[TypeOfStep] = {
                  tsSTART, tsGOTO, tsTAKE, tsPUT_DOWN, tsEND, tsWinner}
# TOS = dict(
#     tsSTART    = tsSTART,   # Startovací krok s prázdným názvem
#     tsGOTO     = tsGOTO,    # Úspěšný přesun do sousedního prostoru
#     tsTAKE     = tsTAKE,    # Úspěšné zvednutí objektu v prostoru
#     tsPUT_DOWN = tsPUT_DOWN,# Úspěšné položení objektu z batohu
#     tsEND      = tsEND,     # Příkaz okamžitě ukončující hru
#     tsWinner   = tsWinner,  # Úspěšné ukončení hry
# )


###########################################################################q


# @dbg.prSEda()
def convert_game(solution:'module'):
    """Od zadaného modulu získá potřebné informace a převede je
    na záznamy do jednotlivých tabulek.
    """
    author_id       = solution.AUTHOR_ID
    game_name       = solution.AUTHOR_ORIG_NAME
    game_description= solution.description()
    game_scenario   = solution.scenario()

    game = Rec_Game(author_id, game_name, game_description)
    game_id = id_game.id
    tbl_Games.add(game)

    step_types:set[TypeOfStep] = set()
    places: dict[str,int] = {} # Pro propojení se sousedy potřebujeme znát ID
    # Seznam propojení prostorů, cíl ještě nemusí existovat, použiju název
    p2p: list[tuple[int,str]] = []       # Do propojovací tabulky sousedů
    step:ScenarioStep

    #-----------------------------------------------------------------------
    # @dbg.prSEd()
    def add_action(step:ScenarioStep, type:TypeOfStep):
        """Zařadí do tabulek poprvé použitou akci."""
        step_types.add(type)
        command = step.command
        message = step.message
        if type._param_count > 0:
            words = command.split()
            command = words[0]
            message = message.removesuffix(words[1])
        record = Rec_Action(game_id, type.name, command, message)
        tbl_Actions.add(record)
    #-----------------------------------------------------------------------
    def add_items(place_id:int, items:tuple[str]):
        """Zařadí do tabulek položky v prostoru, resp. batohu."""
        for item in items:
            # Nepřenositelné předměty budou mít váhu -1
            weight = -1 if item[0]=='#' else 1
            name = item[1:]
            record = Rec_Item(place_id, name, weight)
            tbl_Items.add(record)
    #-----------------------------------------------------------------------
    def add_place(name:int, description:str,
                  items: tuple[str], neighbors:tuple[str]):
        """Zařadí do tabulek prostor (resp. batohu) a jeho položky,
        Připraví závěrečné propojování sousedů.
        """
        record = Rec_Place(game_id, name, description)
        tbl_Places.add(record)
        places[name] = id_place.id
        add_items(id_place.id, items)
        if neighbors:   # Jinými slovy: není to batoh
            nonlocal p2p
            for n in neighbors:
                p2p.append((id_place.id, n))
    #-----------------------------------------------------------------------

    for step in game_scenario:
        type = step.typeOfStep

        if (type in checked_types) and not (type in step_types):
            add_action(step, type)  # Zařazení poprvé použité akce

        if type == tsSTART:     # Na počátku udělám z batohu prostor
            # A přidám do tabulek jeho h-objekty
            add_place('', 'BAG', step.bag, None)

        # Zařazení případného poprvé navštíveného prostoru
        place = step.place
        if not (place in places):
            lines = step.message.split('\n')
            add_place(place, lines[-1], step.items, step.neighbors)

    # Na závěr musíme propojit všechny sousedy
    for source_id, dest_name in p2p:
        dest_id = places[dest_name]
        tbl_Neighbors.add(Rec_Neighbor(source_id,dest_id))



###########################################################################q
dbg.stop_pkg(1, __name__)
