#Příliš žluťoučký kůň úpěl ďábelské ó - PŘÍLIŠ ŽLUŤOUČKÝ KŮŇ ÚPĚL ĎÁBELSKÉ Ó
#L:/66_INP/int2lang.py
"""
Definuje funkce pro převod čísel na slovní vyjádření v zadaném jazyce.
"""
import dbg; dbg.start_mod(1, __name__)
###########################################################################q
# Globální konstanty

__all__ = 'int2 int2cz int2en int2bcz int2ben languages'

from collections import namedtuple

Names = namedtuple('Names', 'language shortcut max_exp '
                   'units tens hundreds orders sextets gender')
Names.__doc__ = """\
Sada konstant potřebných pro převod čísel nezávisle na použitém jazyce:
language - Název rodného (tj. cílového) jazyka zapsaný v daném jazyce
shortcut - Mezinárodní dvoupísmenná zkratka označující cílový jazyk
max_exp  - Řád (mocnina deseti) nejmenšího čísla, které již nejde převést
units    - Názvy čísel do dvaceti
tens     - Názvy desítek
hundreds - Názvy stovek
orders   - Názvy řádů
sextets  - Předpony šestičíslí
gender   - Dodržení rodu
"""

def init_for(lang:str):
    """Nastaví konstanty pro převod čísel do slovního vyjádření v jazyce,
    jehož dvoupísmenná mezinárodní zkratka byla zadána v argumentu.
    """
    par = params.get(lang.lower())
    if not par:
        raise ValueError(f'Není podporován převod pro jazyk {lang!r}')
    global LANGUAGE,SHORTCUT,MAX,UNITS,TENS,HUNDREDS,ORDERS,SEXTETS,GENDER
    LANGUAGE = par.language     # Název jazyka zapsaný v daném jazyce
    SHORTCUT = par.shortcut     # Mezinárodní dvoupísmenná zkratka jazyka
    MAX_EXP  = par.max_exp      # Řád nejmenšího čísla, jež nejde převést
    MAX      = int(10**MAX_EXP) # Nejmenší číslo, jež nejde převést
    UNITS    = par.units        # Názvy čísel do dvaceti
    TENS     = par.tens         # Názvy desítek
    HUNDREDS = par.hundreds     # Názvy stovek
    ORDERS   = par.orders       # Názvy řádů
    SEXTETS  = par.sextets      # Předpony šestičíslí
    GENDER   = par.gender       # Dodržení rodu


params = {
'cz' :  Names(language='Česky', shortcut='cz', max_exp=240,
    units = (
        '',        'jedna',    'dva',       'tři',      'čtyři',
        'pět',     'šest',     'sedm',      'osm',      'devět',
        'deset',   'jedenáct', 'dvanáct',   'třináct',  'čtrnáct',
        'patnáct', 'šestnáct', 'sedmnáct',  'osmnáct',  'devatenáct'
    ),
    tens = (
        '',        '',         'dvacet',    'třicet',   'čtyřicet',
        'padesát', 'šedesát',  'sedmdesát', 'osmdesát', 'devadesát'
    ),
    hundreds = (
        '',        'sto',      'dvě stě',   'tři sta',  'čtyři sta',
        'pět set', 'šest set', 'sedm set',  'osm set',  'devět set'
    ),
    orders = (
        ( '',         '',        ''),
        ( 'tisíc',    'tisíce',  'tisíc'  ),
        ( 'ilion',    'iliony',  'ilionů' ),
        ( 'iliarda',  'iliardy', 'iliard' )
    ),
    sextets = ('',
        #1/6         2/7            3/8            4/9              5/10
        'm',         'b',           'tr',          'kvadr',         'kvint',
        'sext',      'sept',         'okt',        'non',           'dec',
        'undec',     'duodec',       'tredec',     'kvadrdec',      'kvindec',
        'sexdec',    'septendec',    'octodec',    'novemdec',      'vigint',
        'unvigint',  'duovigint',    'trevigint',  'kvadrvigint',   'kvinvigint',
        'sexvigint', 'septenvigint', 'octovigint', 'novemvigint',   'trigint',
        'untrigint', 'duotrigint',   'tretrigint', 'kvadrtrigint, ','kvintrigint',
        'sextrigint','septentrigint','octotrigint','novemtrigint',  'kvaigint',
    ),
    gender = (
        ('',      '',),
        ('jeden', 'jedna', 'jedno'),
        ('dva',   'dvě',   'dvě')
    )
),
'en' :  Names(language='English', shortcut='en', max_exp=9,
    units = (
        '',        'one',     'two',       'three',    'four',
        'five',    'six',     'seven',     'eight',    'nine',
        'ten',     'eleven',  'twelve',    'thirteen', 'fourteen',
        'fifteen', 'sixteen', 'seventeen', 'eighteen', 'nineteen'
    ),
    tens = (
        '',      '',      'twenty',  'thirty', 'forty',
        'fifty', 'sixty', 'seventy', 'eighty', 'ninety'
    ),
    hundreds = ('',
        'one hundred',  'two hundred', 'three hundred', 'four hundred',
        'five hundred', 'six hundred', 'seven hundred', 'eight hundred',
        'nine hundred'
    ),
    orders = (
        ( '',         '',          ''),
        ( 'thousand', 'thousand', 'thousand'  ),
        ( 'illion',   'illion',   'illion'    )
    ),
    sextets = ('',
        #1/6         2/7            3/8            4/9              5/10
        'm',
    ),
    gender = (
        ('',      ''),
        ('one', 'one'),
        ('two', 'two')
    )
)}




############################################################################

# @dbg.prSEd()
def int2(lang:str, number:int, binary=False) -> str|bytes:
    """Převede zadané číslo na slovní vyjádření v zadaném jazyce.

    Argumenty:
    lang:   Dvoupísmenná mezinárodní zkratka definující cílový jazyk
    number: Převáděné číslo, které musí být menší než konstanta MAX,
            jež je součástí interních tabulek pro daný jazyk
    binary: Zda má být vracen bajtový string
    """
    init_for(lang)
    if (number < 0)  or  (MAX <= number):
        raise Exception(f'Lze převádět pouze nezáporná čísla do {MAX}')
    if   number == 0:  result = "nula"          #==========>
    elif number < 20:  result = UNITS[number]   #==========>
    else:
        def trios(number:int) -> list[int]:
            """Rozdělí celé číslo na jednotlivé trojice."""
            result = []
            while True:
                result.append(number % 1000)
                result //= 1000
                if result == 0: break

        # @dbg.prSEdar()
        def sextet(trio:int, order:int) -> str:
            """Vrátí název řádu zadaného trojčíslí."""
            if trio == 0: return ''     #==========>
            row = (order if order < 2 else     # Jednotky a tisíce
                   3     if order % 2 else     # Nedělitelný 2 => jarda
                   2)                          # Dělitelný   2 => lion
            col = (0 if trio==1 else
                   1 if trio<=4 else 2)
            # dbg.prIN(1, f'{order=}, {row=}, {col=}')
            return ' ' + SEXTETS[order // 2] + ORDERS[row][col]

        # @dbg.prSEdar()
        def convert(trio:int, order:int):
            """Převede zadané trojčíslí na string."""
            if trio == 0: return ''     #==========>
            if (trio < 3)  and  order > 0:
                res = GENDER[trio][1 if (order & 1)  and  (order > 1) else
                                   0] # <- liony a tisíce; ^- jardy
                return res  #======>
            units    = trio %  10   # Počet jednotek
            tens     = trio // 10   # Celkový počet desítek
            hundreds = tens // 10   # Počet pouhých stovek
            tens     = tens %  10   # Počet pouhých desítek
            return ( HUNDREDS[hundreds]
                   + (' '  if (hundreds > 0)  and  (number % 100 > 0)  else '')
                   + (UNITS[10*tens + units]  if tens < 2  else
                     (TENS[tens] + (' ' if (units > 0)  else '')
                                 + UNITS[units]) )
               )

        trios = []
        order = 0
        while number > 0:
            trio = number % 1000
            trios.append(convert(trio=trio, order=order)
                        + sextet(trio=trio, order=order))
            number //= 1000
            order  +=  1
        for i in range(len(trios)-1,-1,-1):
            if not trios[i]:  del trios[i]
        result = (' '.join(trios[::-1])).strip()

    if binary:  result = bytes(result, 'utf8')
    return result



###########################################################################q
(languages := [name for name in params]).sort()
# languages.__doc__ = 'Abecedně seřazený seznam zkratek podporovaných jazyků'
# TODO: definovat jako vlastnost

int2cz  = lambda num: int2('cz', num)
int2cz.__name__ = 'int2cz'
int2cz.__doc__  = 'Převod celého čísla na text v češtině'

int2en  = lambda num: int2('en', num)
int2en.__name__ = 'int2en'
int2en.__doc__  = 'Převod celého čísla na text v angličtině'

int2bcz = lambda num: int2('cz', num, True)
int2bcz.__name__ = 'int2bcz'
int2bcz.__doc__  = 'Převod celého čísla na bajtový string v češtině'

int2ben = lambda num: int2('en', num, True)
int2ben.__name__ = 'int2ben'
int2ben.__doc__  = 'Převod celého čísla na bajtový string v angličtině'



###########################################################################q
dbg.stop_mod(1, __name__)
