#Příliš žluťoučký kůň úpěl ďábelské ó - PŘÍLIŠ ŽLUŤOUČKÝ KŮŇ ÚPĚL ĎÁBELSKÉ Ó
#L:/66_INP/m12__Čtení_a_zápis_dat.py
"""
Příkazy zadávané ve výpisech kapitoly:
12  Čtení a zápis dat
"""
import dbg; dbg.start_mod(1, __name__)
###########################################################################q



#Výpis 12.1:	Demonstrace použití metod pro čtení
###########################################################################q
# Příprava pomocných proměnných
import io
text = 'První řádek\nDruhý řádek\nTřetí řádek'
code = b'First line\nSecond line\nThird line'   # Pouze ASCII znaky
sio = io.StringIO(text)     # Znakový vstupně-výstupní proud
bio = io.BytesIO (code)     # Binární vstupně-výstupní proud
# Zadání počtu požadovaných řádků nezaručuje, že se opravdu přečtou
sio.readlines(3),   bio.readlines(3)    # Přečte nejvýše tři řádky
sio.read(6),        bio.read(6)         # Přečte dalších 6 znaků/bajtů
# Ukázka načtení binárních dat přímo do vyhrazené paměti
ba = bytearray(b'Start - End')          # Připraví proměnný bajtový objekt
bio.readinto(ba),   ba                  # Načte bajty do vyhrazené paměti
sio.read(),         bio.read()          # Přečte zbytek
sio.close();        bio.close()         # Na závěr proudy zavře
# Demonstrace proudu jako iterovatelného objektu
def print_lines(stream:'BaseIO') -> None:
    """Vytiskne očíslované řádky daného proudu."""
    for index, line in enumerate(stream): print(f'{index}. {line!r}')

print_lines(io.StringIO(text))  # Argumentem je nově vytvořený proud
print_lines(io.BytesIO(code))
# Pro demonstraci funkce peek() připravíme obyčejný soubor na disku
from pathlib import Path    # Příští příkaz naplní soubor obsahem
(p := Path('xb.txt')).write_bytes(b'Start - Middle - End'), p.read_bytes()
bx = p.open('br')   # Metoda peek() nezaručuje, kolik toho z proudu načte
print(f'{bx.peek(5)=},    {bx.read()=}')
bx.close()  # Nesmíme zapomenout proud zavřít, aby neblokoval prostředky



#Výpis 12.2:	Demonstrace použití metod pro zápis
###########################################################################q
text = ['První řádek\n',  'Druhý ne-řádek',  'Třetí dvojřádek\nKONEC']
code = [b'First line\n', b'Second no-line', b'Third double line\nEND']
sout = io.StringIO();     bout = io.BytesIO()    # Vytvoření proudů
sout.writelines(text),    bout.writelines(code)  # Zápis seznamu řádků
print(f'{sout.getvalue() = }\n{bout.getvalue() = }')
sout.write('_Dodatek'),   bout.write(b'_Appendix')  # Další zápis
print(f'sout: {sout.getvalue()!r}\nbout: {bout.getvalue()}')
sout.close();       bout.close()    # Závěrečné zavření obou proudů



#Výpis 12.3:	Demonstrace funkce a použití splachování
###########################################################################q
# Definujeme cestu a ověříme, že soubor neexistuje (=> nikdo jej neblokuje)
pf = Path('xf.txt');   pf.unlink(True);   pf.exists()   # Neexistuje
# Otevře se proud s implicitní vyrovnávací pamětí a vytvoří se prázdný soubor
fw = pf.open('w');     pf.exists(),     pf.read_text()  # Otevření - prázdný
fw.write('Start\n'),   pf.read_text()   # Data jsou v proudu, ne v souboru
fw.flush(),            pf.read_text()   # Spláchnutí = přesun do souboru
fw.write('Dodatek'),   pf.read_text()   # Další data jsou zase jen v proudu
fw.close(),            pf.read_text()   # Zavření souboru - nejprve spláchne
# Vytvoříme proud s vyrovnávací pamětí na jeden řádek a vyčistíme soubor
fw = pf.open('w', 1);  pf.read_text()   # Soubor je opět prázdný
fw.write('Start'),     pf.read_text()   # Zapsáno do proudu, ne do souboru
fw.write('Stop\nxx'),  pf.read_text()   # Znak '\n' zařídí spláchnutí
fw.write('Dodatek'),   pf.read_text()   # Není znak '\n' => není spláchnuto
fw.close(),            pf.read_text()   # Spláchne se až při zavření souboru



#Výpis 12.4:	Demonstrace použití konstrukce with
###########################################################################q
pt = Path('xt.txt');         pb = Path('xb.txt')
with pt.open('w') as ft,     pb.open('bw') as fb:
    t = ft.write('Zapsaný text zatím není spláchnutý')
    b = fb.write(b'The written text is not yet flushed')
    print(f'Uvnitř:   {t=}, {pt.read_text()=}  -  {b=}, {pb.read_bytes()=}\n'
          f'Uzavření: {ft.closed=},            {fb.closed=}')

print(f'Po ukončení příkazu with: {ft.closed=}, {fb.closed=}\n'
      f'{pt.read_text() = }\n{pb.read_bytes()=}')



#Výpis 12.5:	Pomocná funkce fill() určená pro zaplnění řádků jejich čísly
###########################################################################q
from int2lang import int2cz, int2ben

def fill(path:str|Path, start:int, stop:int, bin:bool=False) -> Path:
    """Vytvoří soubor, v jehož řádcích budou čísla v zadaném rozsahu
    řečená slovy - u textových souborů česky, u binárních anglicky."""
    nl     = b'\n' if bin else '\n'         # Konec řádku podle typu
    fun    = int2ben if bin else  int2cz    # Funkce generující texty
    with open(path, 'w'+('b' if bin else 't')) as stream:
        for i in range(start, stop):
            stream.write(fun(i) + nl)   # Zapíše čísla slovy



#Výpis 12.6:	Demonstrace práce s datovým kurzorem
###########################################################################q
from pathlib  import Path
pt = Path('xnumt.txt');         pb = Path('xnumb.bin')
fill(pt, 1, 10);                fill(pb, 1, 10, 1)  # Vytvoření souborů
to = [0];                       bo = [0]  # Seznamy "offsetů" začátků řádků
# Protože editory číslují řádky od 1, dali jsme do nultého prvku falešný
with pt.open('tr+') as ft,      pb.open('br+') as fb:
    print(f'{ft = }\n{fb = }')
    for i in range(10):     # Zapamatuje si pozice jednotlivých řádků
        to.append(ft.tell());   bo.append(fb.tell())
        t=ft.readline();        t=fb.readline()     # Aby se nevypisovaly
    print(f'Pozice v textovém: {to}\nPozice v bajtovém: {bo}')
    ft.seek(to[6]),             fb.seek(bo[6])      # Najedu zpět na šesté
    ft.readline(),              fb.readline()       # Přečtu je
    ft.write('SEDM'),           fb.write(b'SEVEN')  # Přepíšu následující
    ft.seek(to[7]),             fb.seek(bo[7])      # Zpět na jejich začátky
    ft.readline(),              fb.readline()       # A přečtu nový text



#Výpis 12.7:	Prověrka chování textového proudu
###########################################################################q
with pt.open('r+') as ft:
    ft.seek(to[10]),        ft.readline()
    ft.seek(to[7]),         ft.write('SEDM')
    ft.seek(to[7]),         ft.readline()



#Výpis 12.8:	Demonstrace tvorby ZIP-souborů a práce s nimi
###########################################################################q



#Výpis 12.0:	Popisek
###########################################################################q
"""
Je uloženo v souboru
#L:/66_INP/xxx.py
"""
##### Kapitola 12.0:	Název
#--------------------------------------------------------------------------q
                                                                       #SYNC
###########################################################################q
dbg.stop_mod(1, __name__)
