#Příliš žluťoučký kůň úpěl ďábelské ó - PŘÍLIŠ ŽLUŤOUČKÝ KŮŇ ÚPĚL ĎÁBELSKÉ Ó
#L:/66_INP/m15__JSON.py
"""
Příkazy zadávané ve výpisech kapitoly:
15  Dokumenty ve formátu JSON
"""
import dbg; dbg.start_mod(1, __name__)
###########################################################################q



#Výpis 15.1:	Ukázka převodu vzorových dat do formátu JSON
###########################################################################q
import json     # Následují data se všemi akceptovatelnými typy hodnot
data = {'konst':(None, True, False), 'čísla':(0xFF, 314e-2),
        'divná':[float('nan'), float('inf'), float('-inf')],
        'text': 'String s "uvozovkami", odřádkováním\n'
                "a 'apostrofy'."}
# Implicitně je vše převedeno na ASCII a vráceno jako jeden dlouhý string
print(json.dumps(data))     # Tiskne implicitní podoba převedených dat
# Nebude se převádět na ASCII, budou se řádkovat a odsazovat seřazená data
print(json.dumps(data, ensure_ascii=False, indent=4, sort_keys=True))



#Výpis 15.2:	Reakce systému na nestandardní typ
###########################################################################q
from game.a_01 import steps
json.dumps(steps)



#Výpis 15.3:	Použití vlastního kodéru na převod instancí typu TypeOfStep
###########################################################################q
def convert(obj) -> str:
    """Převede instance typu `TypeOfStep` do formátu JSON tak,
    aby je bylo možno následně opět převést zpět."""
    result = {}
    from game.api.scen_types import TypeOfStep
    if isinstance(obj, TypeOfStep):
        result['__class__']  = obj.__class__.__name__
        result['__module__'] = obj.__class__.__module__
        result['name']       = obj.name
    else:
        raise Exception(f'Objekt neznámého typu: {type(obj)}')
    return result

print(json.dumps(steps[1], default=convert, ensure_ascii=False, indent=4))



#Výpis 15.4:	Použití alternativního postupu pro převod instancí typu TypeOfStep
###########################################################################q
from game.api.scen_types import TypeOfStep
def python_2_json(self):   # Definuje funkci, kterou pak uloží jako metodu
    """Budoucí instanční metoda typu `TypeOfStep` - převede instanci
    na slovník uchovávající název mateřské třídy a jejího modulu
    a hodnoty klíčových atributů."""
    return {'__class__': 'TypeOfStep', '__module__':TypeOfStep.__module__,
            'name': self.__name__}

TypeOfStep.python_2_json = python_2_json  # Uloží ji jako metodu třídy ...
del python_2_json        # TypeOfStep, načež může pomocnou proměnnou smazat

def convert_2_json(obj):
    """Převod objektu Pythonu na objekt formátu JSON, tj. na slovník."""
    if ('python_2_json' in obj.__class__.__dict__
                        and callable(obj.python_2_json)):
        # Objekty definující metodu python_2_json() se převedou samy
        return obj.python_2_json()     # Vracím obdržený objekt formátu JSON
    else:   # Pokus o převod objektů jiných typů vyvolá výjimku
        raise Exception(f'Objekty typu {type(obj)} se neumějí převést do JSON')

print(json.dumps(steps[1], default=convert_2_json, ensure_ascii=False, indent=4))



#Výpis 15.5:	Zpětný převod dat z formátu JSON na objekty Pythonu
###########################################################################q
data['text'] = 'String'
stored = json.dumps(data, ensure_ascii=False)
loaded = json.loads(stored)
print(f'{stored = !s}\n{data = !s}\n{loaded = !s}\n{(data == loaded) = }')
f"{float('nan')==float('nan') = },     {data['divná'][0] = }"
from copy import deepcopy     # Potřebuji skutečné kopie
data2 = deepcopy(data);       del data2['divná'][0]    # Vyhazuji NaN
load2 = deepcopy(loaded);     del load2['divná'][0]
load2['konst'] = tuple(load2['konst']);   load2['čísla'] = tuple(load2['čísla'])
data2 == load2



#Výpis 15.6:	Program pro zpětný převod kroků scénáře z formátu JSON na objekt Pythonu
###########################################################################q
def convert_2_python(d) -> object:
    """Převod objektů JavaScriptu na objekty Pythonu.
    Převáděný objekt obdržíme jako slovník."""
    if '__class__' in d:
        class_name  = d.pop('__class__')    # Odebere ze slovníku třídu
        module_name = d.pop('__module__')   # Odebere ze slovníku modul
        from importlib import import_module
        module = import_module(module_name) # Importuje příslušný modul
        cls    = getattr(module, class_name)# Získá odkaz na objekt třídy
        if ('json_2_python' in cls.__dict__):
            # Objekty definující metodu python_2_json() se převedou samy
            return cls.json_2_python(d)  # Vrací vytvořený objekt Pythonu
        else:  # Pokus o převod objektů jiných typů vyvolá výjimku
            raise Exception(
                f'Objekty typu {module_name}.{class_name}\n'
                f'se neumějí převést z podoby v JSON do Pythonu')
    else:
        raise Exception(
            f'Obdržený objekt ve formátu JSON neposkytuje informace\n'
            f'pro jednoznačný převod do Pythonu\n{d}')

def json_2_python(d) -> object:
    """Převádí prvek s informací o typu kroku na objekt typu `TypeOfStep`."""
    name = d['name']    # Zjistí název daného stylu kroku
    from game.api.scen_types import TypeOfStep
    result = TypeOfStep.get(name)
    return result

# Uloží ji jako metodu třídy TypeOfStep, načež může pomocnou proměnnou smazat
from game.api.scen_types import TypeOfStep
TypeOfStep.json_2_python = json_2_python;      del json_2_python

jsn = json.dumps(steps[1], default=convert_2_json, ensure_ascii=False)
pyt = json.loads(jsn, object_hook=convert_2_python)
print(pyt)



#Výpis 15.7:	Druhá etapa zpětného převodu kroků scénáře
###########################################################################q
from game.api.scenario import ScenarioStep
def pyt_2_steps(py1:list) -> tuple[ScenarioStep]:
    """Dokončí převod objektu ve formátu json na objekt Pythonu.
    Argumentem je výsledek 1. etapy převodu - seznam seznamů prvků kroku."""
    py2 = []
    for lst in py1:
        index     = lst[0]
        typeOfStep= lst[1]
        command   = lst[2]
        message   = lst[3]
        place     = lst[4]
        neighbors = tuple(lst[5])
        items     = tuple(lst[6])
        bag       = tuple(lst[7])
        py2.append(ScenarioStep.new(index, typeOfStep, command, message,
                                    place, neighbors,  items,   bag))
    return tuple(py2)

# Převod kroků scénáře do formátu JSON
from game.a_01 import steps     # Pro jistotu importuji ještě jednou
jsn = json.dumps(steps, default=convert_2_json, ensure_ascii=False, indent=4)
# Zpětný převod
py1 = json.loads(jsn, object_hook=convert_2_python) # 1. etapa zpětného převodu
py2 = pyt_2_steps(py1)                              # 2. etapa zpětného převodu
print(f'{type(py1[0])=},  {type(py2[0])=}\n{py2[1]}', end='')
steps == py2



#Výpis 15.0:	Popisek
###########################################################################q
"""
Je uloženo v souboru
#L:/66_INP/xxx.py
"""
##### Kapitola 15.0:	Název
#--------------------------------------------------------------------------q
                                                                       #SYNC
###########################################################################q
dbg.stop_mod(1, __name__)
