#Příliš žluťoučký kůň úpěl ďábelské ó - PŘÍLIŠ ŽLUŤOUČKÝ KŮŇ ÚPĚL ĎÁBELSKÉ Ó
#L:/66_INP/m16a_Convert.py
"""
Dokumentační komentář modulu.
"""
import dbg; dbg.start_mod(1, __name__)
###########################################################################q

# from dataclasses import *
#
# @dataclass
# class Record:
#     """Záznamy s charakteristikou scénáře."""
#     ID:           str   # Identifikátor scénáře
#     steps:        int   # Počet kroků
#     cmd_chars:    int   # Počet znaků v příkazech
#     message_words:int   # Počet slov v odpovědích
#     places:       int   # Počet prostorů
#     multi_items:  int   # Předmětů uvedených v prostorech byť opakovaně
#     neighbors:    float # Průměrný počet sousedů v kroku
#
#     def to_list(self):
#         """Převede pojmenovanou n-tici na seznam."""
#         return [self.ID, self.steps, self.cmd_chars, self.message_words,
#                 self.places, self.multi_items, self.neighbors]



from collections import namedtuple
Record = namedtuple('Record', 'ID steps cmd_chars message_words '
                              'places multi_items neighbors')
Record.__doc__ = """\
    Alternativní řešení prostřednictvím pojmenovaných n-tic
    usnadňuje některé operace, protože n-tice jsou iterovatelné.

    Význam jednotlivých atributů:
    ID:           str   # Identifikátor scénáře
    steps:        int   # Počet kroků daného scénáře
    cmd_chars:    int   # Počet znaků v příkazech
    message_words:int   # Počet slov v odpovědích
    places:       int   # Počet navštívených prostorů
    multi_items:  int   # Počet zmínek o nějakém h-objektu v prostorech
    neighbors:    float # Průměrný počet sousedů aktuálního prostoru v kroku
    """


# from game.api.scenario import ScenarioStep    # Kvůli apostrofům v anotacích
def convert_steps(ID:str, scenario:tuple['ScenarioStep']) -> Record:
    """Analyzuje zadanou posloupnost kroků a vytvoří o ní záznam.
    ID:       Identifikační string daného scénáře
    scenario: N-tice kroků analyzovaného scénáře
    """
    steps = len(scenario)   # Počet kroků scénáře
    cmd_chars    = 0        # Počet znaků v příkazech
    message_words= 0        # Počet slov v odpovědích
    places       = set()    # Různé prostory (množina zamezí opakování)
    multi_items  = 0        # Počet zmínek o nějakém h-objektu v prostorech
    neighbors    = 0        # Celkem uvedených sousedů
    for step in scenario:
        cmd_chars    += len(step.command)
        message_words+= len(step.message.split())
        places.add(step.place.lower())  # lower() zabezpečí shodné názvy
        multi_items  += len(step.items)
        neighbors    += len(step.neighbors)
    places = len(places)    # Zajímavý je jen počet
    neighbors /= steps      # Převede počet na průměr
    return Record(ID, steps, cmd_chars, message_words, places, multi_items,
                  round(neighbors, 2))



def convert_dict(scenarios:dict[str, tuple['ScenarioStep']]) -> list[Record]:
    """Obdrží slovník se scénáři
    a vrátí seznam záznamů s jejich charakteristikami.
    """
    result = []
    for ID, steps in scenarios.items():
        result.append(convert_steps(ID, steps))
    return result



def write_csv(path, records:list[Record]):
    """Zapíše do zadaného souboru obsah zadaného seznamu záznamů.
    """
    import csv
    with open(path, 'w') as stream:
        writer = csv.writer(stream)
        names  = Record._fields     # Atribut _fields uchovává názvy prvků
        writer.writerow (names)     # Zapíšeme hlavičku
        writer.writerows(records)   # Zapíšeme jednotlivé záznamy



###########################################################################q
dbg.stop_mod(1, __name__)
